package com.monnit.mine.MonnitMineAPI;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.nio.charset.Charset;
import java.security.SecureRandom;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.List;
import java.util.TimeZone;

import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;

import com.monnit.mine.BaseApplication.BitConverter;
import com.monnit.mine.BaseApplication.ExtensionMethods.Extensions;
import com.monnit.mine.Encryption.GW_Encryption;
import com.monnit.mine.Encryption.GW_Encryption.EncResult;
import com.monnit.mine.EventHandling.Event;
import com.monnit.mine.EventHandling.EventListener;
import com.monnit.mine.EventHandling.ExceptionHandlerThread;
import com.monnit.mine.EventHandling.GatewayMessageHandlerThread;
import com.monnit.mine.EventHandling.GatewayResponseHandler;
import com.monnit.mine.EventHandling.PersistGatewayHandlerThread;
import com.monnit.mine.EventHandling.PersistSensorHandlerThread;
import com.monnit.mine.EventHandling.SensorMessageHandlerThread;
import com.monnit.mine.Exceptions.ArgumentNullException;
import com.monnit.mine.MonnitAPN.MSAPI;
import com.monnit.mine.MonnitAPN.MNTEventArgs.*;
import com.monnit.mine.MonnitMineAPI.enums.*;
import com.monnit.mine.WebPackages.IPEndPoint;
import com.monnit.mine.WebPackages.UdpClient;

/// <summary>
/// Class that manages communication with gateways and parses communications to and from sensors.
/// </summary>
public class MineServer
{
    //#region Constructors
    /// <summary>
    /// Initializes a new instance of the Server class.
    /// </summary>
    /// <param name="protocol">Protocol server will accept data from.</param>
    /// <param name="address">IP address to bind socket(s) to.</param>
    /// <param name="port">Port to bind socket(s) to.</param>
    public MineServer(eMineListenerProtocol protocol, InetAddress address, int port) throws IOException
    {
        switch (protocol)
        {
            case TCPAndUDP:
                //TCPSocket = new MineTCPSocket(this, address, port);
            	TCPSocket = new MineTCPSocket(this, address, port);
                UDPSocket = new MineUDPSocket(this, port);
                break;
            case TCP:
                TCPSocket = new MineTCPSocket(this, address, port);
                break;
            case UDP:
                UDPSocket = new MineUDPSocket(this, port);
                break;
        }
        
        gatewayresponsehandler = new GatewayResponseHandler();
        gatewayresponsehandler._Server = this;
        addGatewayResponseHandler(gatewayresponsehandler);
    }
    //#endregion
    
    //#region Properties
    public HashMap<Integer, Gateway> _Gateways = new HashMap<Integer, Gateway>();
    public HashMap<Integer, Sensor> _Sensors = new HashMap<Integer, Sensor>();

    private MineTCPSocket TCPSocket = null;
    private MineUDPSocket UDPSocket = null;
    
    private GatewayResponseHandler gatewayresponsehandler;
    
    
    
    //#endregion

    //#region Public Methods
    /// <summary>
    /// This call starts the socket listeners that the gateways will use to communicate
    /// </summary>
    /// <exception cref="Exception">thrown if IP/Port is already bound to another socket.</exception>
    public void StartServer() throws Exception
    {
    	  	
    	Extensions.debug("Reached StartServer.");
        if (TCPSocket != null)
            TCPSocket.StartTCPListeners();

        if (UDPSocket != null)
            UDPSocket.StartUDPListeners();
        
        Extensions.debug("Finished StartServer.");
    }
    
    /// <summary>
    /// This call stops the socket listeners that the gateways will use to communicate
    /// </summary>
    public void StopServer() throws InterruptedException, IOException
    {
        if (TCPSocket != null)
        {
            TCPSocket.StopTCPListeners();
        }

        if (UDPSocket != null)
        {
            UDPSocket.StopUDPListeners();
        }
    }


    /// <summary>
    /// Adds the gateway to the servers list of valid objects
    /// </summary>
    /// <param name="gateway">Gateway object</param>
    /// <exception cref="ArgumentNullException">Parameter "gateway" is null</exception>
    /// <exception cref="Exception">There is already a gateway registered with the same gatewayID, or gatewayID property of gateway is not valid (Zero is not a valid gatewayID) </exception>
    public void RegisterGateway(Gateway gateway) throws Exception
    {
        if (gateway == null)
            throw new ArgumentNullException("gateway");
        if (gateway.GatewayID < 1)
            throw new Exception("Invalid GatewayID");
        if (FindGateway(gateway.GatewayID) != null)
            throw new Exception("GatewayID " + gateway.GatewayID + " is already registered");

        _Gateways.put(gateway.GatewayID, gateway);
    }

    /// <summary>
    /// Removes the gateway from the servers list of valid objects
    /// </summary>
    /// <param name="gatewayID">Numeric identifier of the gateway object.</param>
    public void RemoveGateway(int gatewayID)
    {
        _Gateways.remove(FindGateway(gatewayID));
    }

    /// <summary>
    /// Access to the object instances used by the server.  Returns null if gatewayID does not belong to any of the registered gateways.
    /// </summary>
    /// <param name="gatewayID">Numeric identifier of the gateway object.</param>
    /// <returns>The gateway object stored in memory</returns>
    public Gateway FindGateway(int gatewayID)
    {
    	try
    	{
    		
    		return _Gateways.get(gatewayID);
    	}
    	catch(Exception ex)
    	{
    		Extensions.debug(ex.toString());
            OnLogException(ex, "MineServer.FindGateway invalid GatewayID");
    	}
    	
    	//System.out.println("    Looking for " + gatewayID);
    	//for(Gateway g : _Gateways) {
    		//System.out.println("    Looking at " + g.GatewayID);
    		//if(g.GatewayID == gatewayID) {
    			//System.out.println("    Returning gateway " + g.GatewayID);
    			//return g;
    		//}
    	//}
        return null;
    }

   

    /// <summary>
    /// Adds the sensor to the servers list of valid objects
    /// </summary>
    /// <param name="sensor">Sensor object</param>
    /// <exception cref="ArgumentNullException">Parameter "sensor" is null</exception>
    /// <exception cref="Exception">There is already a sensor registered with the same sensorID, or sensorID property of sensor is not valid (Zero is not a valid sensorID) </exception>
    public void RegisterSensor(int gatewayID, Sensor sensor) throws Exception
    {
        if (sensor == null)
            throw new ArgumentNullException("sensor");
        if (sensor.SensorID < 1)
            throw new Exception("Invalid SensorID");
        
        Gateway MineGateway = FindGateway(gatewayID);
        if (MineGateway == null)
            throw new Exception("GatewayID "+ gatewayID +" is not registered");
        
        System.out.println("    MineGateway ID: " + MineGateway.GatewayID);
        if (MineGateway.getAssignedSensors().indexOf(sensor) > 0)
            throw new Exception("SensorID "+ sensor.SensorID +" is already assigned to gatewayID " + gatewayID);

        if (FindSensor(sensor.SensorID) == null)
            _Sensors.put(sensor.SensorID, sensor);

        MineGateway.AssignSensor(sensor.SensorID);
    }

    /// <summary>
    /// Removes the sensor from the servers list of valid objects
    /// </summary>
    /// <param name="sensorID">Numeric identifier of the sensor object.</param>
    public void RemoveSensor(int sensorID) throws InterruptedException
    {
        //Make sure sensor is not assigned to any gateways
        for(Gateway g : _Gateways.values())
            g.RemoveSensor(sensorID);

        //Remove sensor from list
        _Sensors.remove(FindSensor(sensorID));
    }

    /// <summary>
    /// Access to the object instances used by the server.  Returns null if sensorID does not belong to any of the registered gateways.
    /// </summary>
    /// <param name="sensorID">Numeric identifier of the sensor object.</param>
    /// <returns>The sensor object stored in memory</returns>
    public Sensor FindSensor(int sensorID) throws InterruptedException
    {
    	for(Sensor s : _Sensors.values()) {
    		if(s.SensorID == sensorID) {
    			return s;
    		}
    	}
    	Extensions.debug("FindSensor is returning null.");
        return null;
    }
    
    public boolean updateSensorHashMap(Sensor sens)
    {
    	_Sensors.put(sens.SensorID, sens);
    	return false;
    }

    /// <summary>
    /// Queues sensor specific command to be sent down
    /// </summary>
    /// <param name="sensorID">Numerical identifier of sensor to send command</param>
    /// <param name="commandFrame">Frame for command as returned by SensorBaseProfile class</param>
    /// <param name="isUrgent">Will set the urgent flag so that the command is broadcast when received by the gateway</param>
    public void EnqueueCommand(int sensorID, byte[] commandFrame, Boolean isUrgent) throws Exception
    {
        Sensor sens = FindSensor(sensorID);
        if (sens == null) {
            throw new Exception("Sensor " + sensorID + " not found");
        }

       
        sens.addActionControlCommand(commandFrame);
        sens.PendingActionControlCommand = true;
        
        if (isUrgent)
        {
            if (!sens.getActionControlCommandIsUrgent())
                sens.setActionControlCommandIsUrgent(true);

            for(Gateway gate : _Gateways.values())
            {
                if(gate.getAssignedSensors().contains(sensorID))
                    gate.HasUrgentTraffic = true;
            }
        }
    }

    /// <summary>
    /// Clear queue of all commands to be sent to sensor
    /// </summary>
    /// <param name="sensorID">Numerical identifier of sensor to send command</param>
    public void ClearCommandQueue(int sensorID) throws Exception
    {
        Sensor sens = FindSensor(sensorID);
        if (sens == null) {
            throw new Exception("Sensor not found");
        }

        sens.ActionControlCommands.clear();
        sens.ActionControlCommandIsUrgent = false;
    }
    //#region Events
    //#region Events Helpers
    private <E> void RaiseEventHelper(Event<E> eventHandler, Object eventArgs) //where T extends EventArgs
    {
        Event<E> handler = eventHandler;
        EventListener<Integer> el = new EventListener<Integer>() {
			@Override
			public void onNotify(java.lang.Object source, Integer message) {
				// TODO Auto-generated method stub
				
			}
        };
        if (handler != null)
        {
        	handler.dispatch(eventArgs, null); //TODO find out what null should actually be
        	/*
            EventHandler[] eventHandlers = handler.GetInvocationList();
            for (EventHandler currentHandler : eventHandlers)
            {
                EventHandler<T> currentSubscriber = (EventHandler<T>)currentHandler;
                try
                {
                    currentSubscriber(this, eventArgs);
                }
                catch (Exception ex)
                {
                    try
                    {
                        //log exception location of method that called this helper
                        OnLogException(ex, " MineServer." + new StackFrame(1).GetMethod().Name);
                    }
                    catch (Exception e) { }//If logging failed than ignore error.
                }
            }*/
        }
    }

    /// <summary>
    /// Fired when an exception is caught allowing the exception to be logged.
    /// </summary>
    //public Event<HandleLogExceptionEventArgs> LogException;
    /// <summary>
    /// Fires the LogException event.
    /// </summary>
    /// <param name="ex">Exception to be passed.</param>
    /// <param name="location">Method name where the exception occurred.</param>
    protected void OnLogException(Exception ex, String location)
    {
    	HandleExceptionProcessing(ex);
        //RaiseEventHelper(LogException, new HandleLogExceptionEventArgs(ex, location));
    }

    /// <summary>
    /// Fired when changes have been made to the gateway object that can be persisted to a permanent data store.
    /// </summary>
    //public Event<HandlePersistGatewayEventArgs> PersistGateway;
    /// <summary>
    /// Fires PersistGateway event.
    /// </summary>
    /// <param name="gatewayID">Numerical identifier of gateway.</param>
    protected void OnPersistGateway(long gatewayID)
    {
    	HandlePersistGatewayProcessing(gatewayID);
    }

    /// <summary>
    /// Fired when changes have been made to the sensor object that can be persisted to a permanent data store.
    /// </summary>
    //public Event<HandlePersistSensorEventArgs> PersistSensor;
    /// <summary>
    /// Fires PersitSensor event.
    /// </summary>
    /// <param name="sensorID">Numerical identifier of sensor.</param>
    protected void OnPersistSensor(long sensorID)
    {
    	HandlePersistSensorProcessing(sensorID);
    }

    /// <summary>
    /// Fired when a message is received from a gateway.
    /// </summary>
    //public Event<GatewayMessageEventArgs> GatewayMessage;
    /// <summary>
    /// Fires Gateway event.
    /// </summary>
    /// <param name="gatewayMessage">Class holding data about the gateway communication.</param>
    protected void OnGatewayMessage(GatewayMessage gatewayMessage) throws InterruptedException
    {
    	Extensions.debug("Reached OnGatewayMessage", 0);
    	HandleGatewayMessageProcessing(gatewayMessage);
        //RaiseEventHelper(GatewayMessage, new GatewayMessageEventArgs(gatewayMessage));
    }
    
    
    private void OnLogException(Exception e) {
    	HandleExceptionProcessing(e);
    }

    /// <summary>
    /// Fired when one or more sensor messages are received.
    /// </summary>
    //public Event<SensorMessagesEventArgs> SensorMessage;
    /// <summary>
    /// Fires the SensorMessage event.
    /// </summary>
    /// <param name="sensorMessages">List of sensor data messages.</param>
    /// <param name="gatewayID">Numerical identifier of gateway that send the messages.</param>
    protected void OnSensorMessages(List<SensorMessage> sensorMessages, int gatewayID)
    {
    	HandleSensorMessageProcessing(sensorMessages, FindGateway(gatewayID));
        //RaiseEventHelper(SensorMessage, new SensorMessagesEventArgs(sensorMessages, gatewayID));
    }

    
  //  class MineServerInternal
   // {
    	
        //#region Internal Methods
        public Gateway FindGatewayByExternalIdentifier(String externalIdentifier)
        {
        	for(Gateway gate : _Gateways.values()) {
        		if(gate.ExternalIdentifier == externalIdentifier) return gate;
        	}
        	return null;
        }

        private Object threadLock = new Object();
        private List<InboundPacket> ActivePackets = new ArrayList<InboundPacket>();
        private int _RunningPackets = 0;
        public byte[] ProcessBytes(byte[] Message) throws Exception
        {
        	Extensions.debug("ProcessBytes", 50);
        	InboundPacket Packet = new InboundPacket();
        	try
        	{
        		_RunningPackets++;
        		if (Message.length < 16)
                {
                    try
                    {
                        String str = new String(Message).trim(); //Extensions.byteToAsciiString(Message, 0, Message.length); //can convert from stri
                        if (str == "TCPTest")
                        {
                        	Extensions.debug("TCPTest");
                            return  "Success".getBytes(); //Extensions.asciiStringToByte("Success");
                        }
                        else if (str.toLowerCase().startsWith("gwp="))
                        {
                            byte[] ResponseBuffer;
                            int gwID = Integer.parseInt((str.toLowerCase().trim().replace("gwp=", "").replace("/0", "")));
                            Gateway Gateway = FindGateway(gwID);
                            if (Gateway != null)
                            {
                                if (!Gateway.HasUrgentTraffic)
                                    ResponseBuffer = Extensions.asciiStringToByte("OK");
                                else
                                    ResponseBuffer = Extensions.asciiStringToByte("CB");

                                return ResponseBuffer;
                            }
                        }
                    }
                    catch (Exception ex)
                    {
                    	Extensions.debug(ex.toString());
                        OnLogException(ex, "MineServer.ProcessBytes invalid ping");
                    }
                }

        		Extensions.debug("Message.length " + Message.length, 50);
        		
        		Gateway SensorGateway = null;
        		Calendar receivedDate = CurrentUTC();
        		Packet.setReceivedDate(receivedDate); //todo: ad in 1/1/2010
        		Packet.setVersion(Message[8]);
        		Packet.setSequence(Message[9]);

        		//UDP only supported on Gateway Protocol Version 2
        		Extensions.debug("Getting GatewayID.");
        		Packet.setGatewayID(BitConverter.toInt32(Message, 0));
        		
        		
        		
        		if (Packet.getGatewayID() < Integer.MAX_VALUE)
        			SensorGateway = FindGateway(Packet.getGatewayID());
        		
        		Packet.setSecurity(BitConverter.ToUInt32(Message, 4));
        		
        		Extensions.debug("Security: " + Packet.getSecurity(), 50);
        		
        		boolean needsResetEncryption = false;
        		if (Packet.getSecurity() > 0 && SensorGateway != null)
        		{
        			try
        			{
        				Extensions.debug("Not Decrypted Message: " + Extensions.hexStr(Message), 50);
        				Message = DecryptPacket(SensorGateway, Message);
        				Extensions.debug("Decrypted Message: " + Extensions.hexStr(Message), 50);
        				eGatewayMessageType responseType = eGatewayMessageType.ToMessageType(Message[12]); //throws exception if the response type is not valid.
        				if(responseType == null) {
        					//e.printStackTrace();
        					Extensions.debug("Decryption failed, invalid response type.", 50);
        					needsResetEncryption = true;
        				}
        			}
        			catch (Exception e)
        			{
        				//e.printStackTrace();
        				Extensions.debug("Decryption threw exception " + e, 50);
        				needsResetEncryption = true;
        			}
        			
        			if(needsResetEncryption) {
        				return resetEncryptionAck(SensorGateway, Packet);
        			}
        		}
        		else {
        			Extensions.debug("No decryption necessary.", 35);
        		}
        		
        		if(needsResetEncryption) {
        			Extensions.debug("\n\nReturn failed to skip packet.\n\n", 35);
        		}
        		
        		Packet.setPower(BitConverter.ToUInt16(Message, 10));
        		Packet.setResponse(Message[12] & 0xFF);
        		Packet.setMore(Message[13] != 0);
        		Packet.setMessageCount(BitConverter.ToUInt16(Message, 14));
        		Packet.setMessage(Message);
        		
        		Boolean DuplicatePacket = false;
        		//lock (UdpConLock)
        		synchronized  (threadLock)
        		{
        			for(InboundPacket ip : ActivePackets)
        			{

        				if (ip.getGatewayID() == Packet.getGatewayID() && ip.getSequence() == Packet.getSequence())
        				{
        					Extensions.debug("Duplicate packet!", 50);
        					DuplicatePacket = true;
        					Packet.isDuplicate = true;
        					break;
        				}
        			}
        			if (!DuplicatePacket) {
        				Extensions.debug(Packet.toString());
        				ActivePackets.add(Packet);
        			}
        		}

        		Extensions.debug("Duplicate Packet: " + DuplicatePacket);
    			synchronized(threadLock)
    			{
    				ActivePackets.remove(Packet);

    				for (int i = ActivePackets.size(); i > 0; i--)
    				{
    					//if Packet is older than 15 seconds than remove from list and accept next packet
    					//if our UDP response failed we need to accept another packet so we can respond to it again
    					//Assuming that the first packet processed has already updated the gateway so we won't 
    					//process repeat data but just send a repeat response.
    					Calendar c = CurrentUTC();
    		        	c.add(Calendar.SECOND, -15);
    					
    					Date fifteenSecondsAgo = c.getTime();
    					if (fifteenSecondsAgo.after(ActivePackets.get(i - 1).getReceivedDate().getTime()))
    						ActivePackets.remove(i - 1);
    				}
    			}
    			
    			if (Packet.getGatewayID() == Integer.MAX_VALUE && Packet.getSecurity() == 0 && Packet.getResponse() == 7)//Wifi Lookup from MAC
    				SensorGateway = FindGatewayByExternalIdentifier(BitConverter.toString(Packet.getMessage(), 16, 6).replace("-", ""));

    			if (SensorGateway != null)
    			{
    				return ProcessPacket(Packet, SensorGateway);
    			}
        	}
        	catch (Exception ex)
        	{
        		ex.printStackTrace();
        		Extensions.debug("Caught Exception in lowest catch of HandleUDPSocketComm: " + ex.toString());
        		OnLogException(ex, "MineUDPSocket.HandleUDPSocketComm(byte[] Message, IPEndPoint clientEndPoint, UdpClient udpListener)");
        	}
        	finally
        	{
        		_RunningPackets--;
        	}
        	throw new Exception("Exception occurred, or Gateway " + Packet.getGatewayID() + " is not registered.");
        } 

        void ProcessPacket(InboundPacket packet, Gateway sensorGateway, InputStream inputStream, OutputStream outputStream)
        {
            ProcessPacket(packet, sensorGateway, inputStream, outputStream, null, null);
        }

        void ProcessPacket(InboundPacket packet, Gateway sensorGateway, IPEndPoint clientEndPoint, UdpClient udpListener)
        {
            ProcessPacket(packet, sensorGateway, null, null, clientEndPoint, udpListener);
        }
        
        byte[] ProcessPacket(InboundPacket packet, Gateway sensorGateway) throws Exception
        {
        	Extensions.debug("ProcessPacket", 50);
        	
        	Boolean GatewayUpdated = false;
            byte[] Header = ResponseHeader(packet);
            
            Extensions.debug("Header length: " + Header.length, 35);
            
            byte[] ResponseBuffer = null;
            GatewayMessage GM = new GatewayMessage();
            
            Extensions.debug("sensorGatewayID: " + sensorGateway.GatewayID + " isdirty: " + sensorGateway.getIsDirty());

            Boolean DuplicateMessage = packet.getSequence() != 0 && sensorGateway.LastSequence == packet.getSequence();
            if (!DuplicateMessage)
            {
                GM.GatewayID = sensorGateway.GatewayID;
                GM.Power = packet.getPower();
                GM.MessageType = packet.getResponse();

                if (GM.MessageType < 7)
                    GM.setMessageCount(packet.getMessageCount());
                
              Calendar receivedDate =  CurrentUTC();
                GM.ReceivedDate = receivedDate;


                sensorGateway.LastSequence = packet.getSequence();
            }
            else {
            	packet.isDuplicate = true;
            }

            Extensions.debug("RESPONSE CODE "+packet.getResponse(), 50);
            
            Extensions.debug("Message: " + Extensions.hexStr(packet.getMessage()), 35);
            
            
            switch(packet.getResponse())
            {
                case 0:
                	Extensions.debug("Get message, case 0", 6);
                	Extensions.debug("Duplicate message? "+DuplicateMessage, 6);
                    if (!packet.isDuplicate)
                    {
                        ProcessMessages(packet, sensorGateway);
                    }

                    Extensions.debug("Get more? "+ packet.getMore(), 3);
                    if (packet.getMore())
                    {
                        ResponseBuffer = ReturnEmptyAck(Header);
                    }
                    else
                    {
                    	//Extensions.hexDebug(Header);
                        Header = UpdateResponse(Header, 1);
                        Extensions.hexDebug(Header);
                        packet.setMessageCount(0); // Reset MessageCount to 0 so that I send the first down message
                        ResponseBuffer = SendDownMessages(packet, sensorGateway, Header, ResponseBuffer);
                        Extensions.debug("ResponseBuffer: " + ResponseBuffer, 13);
                        if (sensorGateway.HasUrgentTraffic)
                        {
                            sensorGateway.HasUrgentTraffic = false;
                            GatewayUpdated = true;
                        }
                    }
                    break;
                case 1: //Send Sensor messages (Pending requests)
                	Extensions.debug("Case 1", 6);
                    ResponseBuffer = SendDownMessages(packet, sensorGateway, Header, ResponseBuffer);
                    if (sensorGateway.HasUrgentTraffic)
                    {
                        sensorGateway.HasUrgentTraffic = false;
                        GatewayUpdated = true;
                    }
                    break;
                case 2: // Send Sensor ID's to update sensor list
                	Extensions.debug("Case 2", 5);
                    ResponseBuffer = SensorIDList(sensorGateway, packet, Header);
                    break;
                case 3: // Network Configuration has been updated.
                	Extensions.debug("Case 3");
                    sensorGateway.IsDirty = false;
                    ResponseBuffer = ReturnZeroAck(Header);
                    GatewayUpdated = true;
                    break;
                case 4: // Reform Network command was received.
                	Extensions.debug("Case 4");
                    sensorGateway.SendResetNetworkRequest = false;
                    ResponseBuffer = ReturnZeroAck(Header);
                    GatewayUpdated = true;
                    break;
                case 5:// Gateway reset to factory defaults (long press)  Deprecated no longer used since EGW 3.0
                	Extensions.debug("Case 5");
                    ResponseBuffer = ReturnEmptyAck(Header);
                    GatewayUpdated = true;
                    break;
                //TODO: Wi-Fi
                /*
                case 7: //Return Gateway and Sensor ID's from WiFi Sensors Mac Address
                    ResponseBuffer = WIFIInfo(Header, sensorGateway);
                    break;
                */
                case 8: //Update Database to match gateway config
                	//REQUIRES PAYLOAD
                	Extensions.debug("Case 8", 8);
                    ResponseBuffer = SaveGatewayConfiguration(sensorGateway, Header, packet.getMessage(), packet.getVersion());
                    GatewayUpdated = true;
                    break;
                case 10: //Record Status Information (Power usage, etc)
                	Extensions.debug("Case 10 - Empty Ack", 10);
                    ResponseBuffer = ReturnEmptyAck(Header);
                    break;
                case 11: //Record Gateway Startup Information (Version, MAC Address)
                	//Requires Payload
                	Extensions.debug("Case 11");
                    RecordGatewayStartup(sensorGateway, packet);
                    ResponseBuffer = ReturnStartupPacket(sensorGateway, Header);
                    GatewayUpdated = true;
                    Extensions.debug("Finished Case 11");
                    break;
                default:
                	Extensions.debug("Case default", 35);
                   	ResponseBuffer = null;
                   	GatewayUpdated = true;
                   	DuplicateMessage = true;
                    //OnLogException(new Exception("Unrecognized Command: " + packet.getResponse() + "\nSensorServer.ProcessPacket(InboundPacket packet, IGateway sensorGateway, NetworkStream clientStream, IPEndPoint clientEndPoint, UdpClient udpListener)"));
                    //

                    break;
            }
            
            if(GatewayUpdated)
                OnPersistGateway(sensorGateway.GatewayID);

            Extensions.debug("Reached duplicate message check.", 35);
            if (!packet.isDuplicate)
            {
                OnGatewayMessage(GM);
            }
            
        	if (packet.getResponse() != 11 && ResponseBuffer[8] > 1 && BitConverter.toInt32(ResponseBuffer, 4) > 0)//Encrypt Packet 
        	{
        		Extensions.debug("Encrypting packet" , 50);
                ResponseBuffer = EncryptPacket(sensorGateway, ResponseBuffer);
        	}
            
            return ResponseBuffer;
        }
        
        void ProcessPacket(InboundPacket packet, Gateway sensorGateway, InputStream inputStream, OutputStream outputStream, IPEndPoint clientEndPoint, UdpClient udpListener)
        {
            try
            {
            	byte[] ResponseBuffer = ProcessPacket(packet, sensorGateway);

                Extensions.debug("Reached end of processpacket switch with " +Extensions.hexStr(ResponseBuffer), 35);
                if (ResponseBuffer != null)
                {

                	
                    if (outputStream != null)//TCP Communication
                    {
                    	//System.out.println("TCP:\n" + Extensions.hexStr(ResponseBuffer));
                    	outputStream.write(ResponseBuffer, 0, ResponseBuffer.length);
                    	outputStream.flush();
                    	Extensions.debug("Sent ResponseBuffer: " + new String(ResponseBuffer, "ASCII"), 50);
                    }
                    else
                    {
                    	Extensions.debug("Sending response " + Extensions.hexStr(ResponseBuffer), 50);
                        udpListener.SendTo(ResponseBuffer, clientEndPoint);
                        Extensions.debug("Sent response", 50);
                    }
                }
                else {
                	//if failed and udp, allow other udp listener to take the port
                	if(outputStream == null) {
                		udpListener.close();
                	}
                }
            }
            catch (Exception ex)
            {
                OnLogException(ex, "SensorServer.ProcessPacket(InboundPacket packet, IGateway sensorGateway, NetworkStream clientStream, IPEndPoint clientEndPoint, UdpClient udpListener)");
            }
        }
        //#endregion

        //#region Gateway Commands
        
        public boolean resetEncryptionAck(Gateway SensorGateway, InboundPacket Packet, OutputStream outputStream, MineServer ServerInstance) throws Exception {
            // only respond once per hour, if the first 11 message doesn't work let the gateway reset to get the encryption key reset
        	Calendar oneHourAgo = CurrentUTC();
        	oneHourAgo.add(Calendar.HOUR, -1);
        	
            if (SensorGateway.DecryptFailureDate == null || SensorGateway.DecryptFailureDate.before(oneHourAgo))
            {
                //Send an 11 back with the encryption key to reset gateway encryption
            	Extensions.debug("Sending 11 back to reset encryption.", 50);
                Packet.setResponse(11);
                Packet.setSecurity(0);
                byte[] ResponseBuffer = ServerInstance.ReturnStartupPacket(SensorGateway, ServerInstance.ResponseHeader(Packet));
                outputStream.write(ResponseBuffer, 0, ResponseBuffer.length);
                outputStream.flush();

                SensorGateway.setDecryptFailureDate(CurrentUTC());
                return true;
            }
            Extensions.debug("Not sending 11 back to reset encryption.", 30);
            return false;
        }

        public boolean resetEncryptionAck(Gateway SensorGateway, InboundPacket Packet, UdpClient udpListener, MineServer ServerInstance, IPEndPoint clientEndPoint) throws Exception {
            // only respond once per hour, if the first 11 message doesn't work let the gateway reset to get the encryption key reset
        	Calendar oneHourAgo = CurrentUTC();
        	oneHourAgo.add(Calendar.HOUR, -1);
        	
            if (SensorGateway.DecryptFailureDate == null || SensorGateway.DecryptFailureDate.before(oneHourAgo))
            {
                //Send an 11 back with the encryption key to reset gateway encryption
            	Extensions.debug("Sending 11 back to reset encryption.", 50);
                Packet.setResponse(11);
                Packet.setSecurity(0);
                byte[] ResponseBuffer = ServerInstance.ReturnStartupPacket(SensorGateway, ServerInstance.ResponseHeader(Packet));
                udpListener.SendTo(ResponseBuffer, clientEndPoint);

                SensorGateway.setDecryptFailureDate(CurrentUTC());
                return true;
            }
            Extensions.debug("Not sending 11 back to reset encryption.", 30);
            return false;
        }
        
        public byte[] resetEncryptionAck(Gateway SensorGateway, InboundPacket Packet) throws Exception {
            // only respond once per hour, if the first 11 message doesn't work let the gateway reset to get the encryption key reset
        	Extensions.debug("Entered resetEncryptionAck", 50);
        	Calendar oneHourAgo = CurrentUTC();
        	oneHourAgo.add(Calendar.HOUR, -1);
        	
            if (SensorGateway.DecryptFailureDate == null || SensorGateway.DecryptFailureDate.before(oneHourAgo))
            {
                //Send an 11 back with the encryption key to reset gateway encryption
            	Extensions.debug("Sending 11 back to reset encryption.", 30);
                Packet.setResponse(11);
                Packet.setSecurity(0);
                byte[] ResponseBuffer = ReturnStartupPacket(SensorGateway, ResponseHeader(Packet));
                //udpListener.SendTo(ResponseBuffer, clientEndPoint);

                SensorGateway.setDecryptFailureDate(CurrentUTC());
                return ResponseBuffer;
            }
            Extensions.debug("Not sending 11 back to reset encryption.", 30);
            return null;
        }
        
//        /
        
        //#region Command 0 Process Data
        private void ProcessMessages(InboundPacket objPacket, Gateway gateway) throws InterruptedException
        {
        	Extensions.debug("Reached ProcessMessages", 3);
            InboundPacket Packet = objPacket;
            List<SensorMessage> MessageList = new ArrayList<SensorMessage>();
            try
            {
                int MessagesReceived = 0;
                int CurrentIndex = 16;
                Extensions.debug("Reached Extensions.tomajorversion");
                int ProtocolVersion = Extensions.ToMajorVersion(gateway.RadioFirmwareVersion) == 1 ? 1 : 2;
                Extensions.debug("Reached while loop in ProcessMessages.");
                while (MessagesReceived < Packet.getMessageCount() && CurrentIndex < Packet.getMessage().length)
                {
                	Extensions.debug("Cycling through while loop in ProcessMessages, Index " + CurrentIndex, 6);
                    MessagesReceived++;
                    MSAPI Msg = ParseMessage(Packet.getMessage(), Packet.getGatewayID(), ProtocolVersion, CurrentIndex);
                    if (Msg != null)
                    {
                    	CurrentIndex+= Packet.getMessage()[CurrentIndex + 5] + 3 + 4;
                    	ProcessMessage(Msg, MessageList);
                    }
                    else {
                    	CurrentIndex++;
                    }
                }
                
                Extensions.debug("Messages to handle: " + MessageList.size(), 2);
                if(MessageList.size() > 0) {
                	//System.out.println("Calling OnSensorMessages");
                    OnSensorMessages(MessageList, gateway.GatewayID);// Handle Synchronously for Control/Alert capabilities
                }
                //new Thread(() => { OnHandleSensorMessages(MessageList, gateway); }).Start();
                
            }
            catch (Exception ex)
            {
            	Extensions.debug("Caught exception in try of ProcessMessages: " + ex.toString());
                OnLogException(ex, "MineServer.ProcessMessages(InboundPacket packet)");
            }
        }
        
        private MSAPI ParseMessage(byte[] buffer, long APNID, int protocolVersion, int index) throws InterruptedException
        {
        	Extensions.debug("Entering ParseMessage", 2);
            try
            {
                if (buffer[index + 4] == MSAPI.START_FRAME_DELIMINATOR)//Frame Delimiter
                {
                	Extensions.debug("Entering first if", 2);
                    int MessageLength = (int)(buffer[index + 5]) + 3;
                    try
                    {
                        byte[] messagebuffer = new byte[MessageLength];
                        System.arraycopy(buffer, index + 4, messagebuffer, 0, MessageLength);

                        MSAPI msg = new MSAPI(messagebuffer);
                        //Verify Frame CRC here---------------
                        /*
                        if (msg.CalcFrameCRC() != msg.getFrameCRC())
                        {
                        	Extensions.debug("Entering second if", 2);
                            String hex = BitConverter.toString(messagebuffer);
                            hex = hex.replace("-", "");
                            OnLogException(new Exception("CRC failed for packet- " + hex), "MineServer.ParseMessage(byte[] message, int StartIndex, long APNID)");
                            index++;
                            Extensions.debug("Returning null in second if", 2);
                            return null;
                        }*/
                        long time = BitConverter.toLongFromUint32(buffer, index);
                       
                       Extensions.debug("Index: "+index, 24);
                       Extensions.debug("Time: " + time, 24);
                        Calendar cal = GetCalendarFromSeconds(time);
                        
                        msg.setMessageDate(cal);
                        msg.setAPNID(APNID);
                        msg.setProtocolVersion(protocolVersion);
                        index += MessageLength + 4;
                        //Extensions.debug("    Returning actual time", 23);
                        return msg;
                    }
                    catch (Exception ex)
                    {
                    	ex.printStackTrace();
                        OnLogException(ex, "MineServer.ParseMessage(byte[] message, int StartIndex, long APNID) -> Inner || ");
                    }

                }
                Extensions.debug("Not entering first if", 6);
            }
            catch (Exception ex)
            {
            	ex.printStackTrace();
                OnLogException(ex, "MineServer.ParseMessage(byte[] message, int StartIndex, long APNID) || Index("+index+") || APNID("+APNID+") || Message"+buffer.toString()+" || ");
            }

            index++;
            Extensions.debug("Returning null in ParseMessage", 6);
            return null;
        }

        private void ProcessMessage(MSAPI msg, List<SensorMessage> messageList) throws InterruptedException
        {
        	Extensions.debug("    Entering ProcessMessage", 2);
            if (msg.frame[0] == 0xC5 && msg.frame[1] == 0x03 && msg.frame[3] == 0xFF)
            {
                Extensions.debug("    Error nothing to process (for now)", 2);
                return;
            }

            try
            {
                //Look for more traffic to generate
                long id = msg.ExtractCMDAddress();
                Extensions.debug("    Before 'if' that adds msg to list with Command="+msg.Command + " and looking for " + MSAPI.CMD_DATA_MESSAGE + " or " + MSAPI.CMD_DATA_MESSAGE_DL, 2);
                if (msg.Command == MSAPI.CMD_DATA_MESSAGE || msg.Command == MSAPI.CMD_DATA_MESSAGE_DL)
                {
                	Extensions.debug("        Adding SensorMessage to list", 20);
                    SensorMessage SM = new SensorMessage((int)(msg.getDM_SensorID()), msg.getDM_NodeDesc(), msg.getDM_MessageDate(), msg.getDM_State(), msg.getDM_RssiInbound(), msg.getDM_Voltage(), msg.getDM_RawData());
                    messageList.add(SM);
                }
                else
                {
/*
                	msg.WriteDataAck += new Event<WriteConfigEventArgs>(HandleWriteDataAck);
                    msg.ReadDataAck += new Event<ReadConfigEventArgs>(HandleReadDataAck);
                    msg.ACTLAck += new Event<ActionCtrlEventArgs>(HandleActionControlAck);
                    msg.ParentMsg += new Event<ParentMessageEventArgs>(HandleParentMessage);
*/
                	//System.out.println("Non data-message message.");
                    String smsg = msg.RX_SHandler(this);
                    System.out.println(smsg);
/*
                    msg.WriteDataAck -= new Event<WriteConfigEventArgs>(HandleWriteDataAck);
                    msg.ReadDataAck -= new Event<ReadConfigEventArgs>(HandleReadDataAck);
                    msg.ACTLAck -= new Event<ActionCtrlEventArgs>(HandleActionControlAck);
                    msg.ParentMsg -= new Event<ParentMessageEventArgs>(HandleParentMessage);
*/
                }
            }
            catch (Exception ex)
            {
            	Extensions.debug("Caught Exception in ProcessMessage " + ex.toString(), 2);
            	ex.printStackTrace();
                if (msg == null || msg.frame == null)
                {
                    OnLogException(new Exception("MSG is NULL", ex), "MineServer.ProcessMessage(MSAPI msg)");
                }
                else
                {
                    String message = String.format("MSG Frame: %s", (""+msg.frame).replace("-", ""));
                    OnLogException(new Exception(message, ex), "MineServer.ProcessMessage(MSAPI msg)");
                }
            }
        }

        public void HandleWriteDataAck(Object sender, WriteConfigEventArgs e) throws InterruptedException
        {
            Sensor s = FindSensor((int)(e.getSensorID()));
            eConfigSection ecs = eConfigSection.ToConfig(e.getConfigSection());
            switch (ecs)
            {
                case General_1:
                    s.setGeneralConfig1Dirty(false);
                    break;
                case General_2:
                    s.setGeneralConfig2Dirty(false);
                    break;
                case Profile_1:
                    s.setProfileConfig1Dirty(false);
                    break;
                case Profile_2:
                    s.setProfileConfig2Dirty(false);
                    break;
            }
            OnPersistSensor(s.SensorID);
        }

        public void HandleReadDataAck(Object sender, ReadConfigEventArgs e) throws Exception
        {
            Sensor s = FindSensor((int)(e.getSensorID()));
            eConfigSection ecs = eConfigSection.ToConfig(e.getConfigSection().Value());
            s.FillConfig(ecs, e.getData());
            OnPersistSensor(s.SensorID);
        }

        public void HandleActionControlAck(Object sender, ActionCtrlEventArgs e) throws InterruptedException
        {
        	System.out.println("HandleActionControlAck");
        	Gateway g = FindGateway((int) e.getSID()); 
        	if(g != null) {
        		g.PendingActionControlCommand = false;
        		OnPersistGateway(g.GatewayID);
        	}
        	else {
        		Sensor s = FindSensor((int)(e.getSID()));
        		if(s==null) return;
        		s.PendingActionControlCommand = false;
        		if(s.getActionControlCommands().size() > 0)
        			s.ActionControlCommands.remove(0);
        		s.ActionControlCommandIsUrgent = false;
        		OnPersistSensor(s.SensorID);
        	}
        }

        public void HandleParentMessage(Object sender, ParentMessageEventArgs e)
        {
            //e.SensorID
            //e.ParentID
        }
        
        //#endregion

        //#region Command 1 Send Down Messages
        private byte[] SendDownMessages(InboundPacket packet, Gateway sensorGateway, byte[] Header, byte[] ResponseBuffer) throws Exception
        {
        	Extensions.debug("Reached SendDownMessages.", 20);
            ResponseBuffer = GatewayUpdates(sensorGateway, Header, packet.getSecurity());

            Extensions.debug("Reached IF in SendDownMessages", 20);
            if (ResponseBuffer == null) {
                ResponseBuffer = RetrieveMessages(sensorGateway, packet, Header);
            }

            Extensions.debug("Finished SendDownMessages: " + Extensions.hexStr(ResponseBuffer), 20);
            return ResponseBuffer;
        }

        private byte[] GatewayUpdates(Gateway gateway, byte[] header, long apnCSNetID) throws InterruptedException
        {
        	Extensions.debug("Reached GatewayUpdates.\n\tDirty? " + gateway.getIsDirty() + "\n\tReform? " + gateway.SendResetNetworkRequest, 20);
        	if (gateway.getIsDirty())//Send Gateway Configuration
            {
            	Extensions.debug("Gateway is dirty and will be updated.", 20);
                return GetGatewayConfiguration(gateway, header);
            }
            else if (gateway.SendResetNetworkRequest)//Send Reform Network (Sensor Network) Command
            {
            	Extensions.debug("Reached reform network in GatewayUpdates", 20);
                return GetNetworkResetCommand(header);
            }

        	Extensions.debug("Finished GatewayUpdates", 20);
            return null;
        }

        private byte[] GetNetworkResetCommand(byte[] header)
        {
            byte[] responseBuffer;
            // Change the response type to 4
            header[14] = 0x04; //4;

            responseBuffer = new byte[header.length + 2];//18
            System.arraycopy(header, 0, responseBuffer, 0, header.length);

            //More Flag = 0 (false)
            responseBuffer[15] = 0x00;

            return responseBuffer;
        }

        private byte[] GetGatewayConfiguration(Gateway gateway, byte[] header) throws InterruptedException
        {

        	Extensions.debug("Entered GetGatewayConfiguration", 20);
            byte[] responseBuffer;
            // Change the response type to 3
            header[14] = 0x03; //3;

            byte[] HostAddress = ParseIP(gateway.ServerHostAddress);
            byte[] HostAddress2;
            String Password = "";

            switch (gateway.GatewayType)
            {
                default:
                case USB://USB Service Gateway
                    //#region USB Service
                    responseBuffer = new byte[header.length + 2 + 512];//512 + 18
                    System.arraycopy(header, 0, responseBuffer, 0, header.length);
                    //Assemble the Configuration page
                    for (int i = 0; i < 512; i++) responseBuffer[header.length + 2 + i] = (byte) 0xFF;  //Base fill with 0xFF

                    //MNVM_RADIOCHANMASK_32 is at Offset 4 (header.length + 4;
                    System.arraycopy(BitConverter.getBytes((int)(0xFFFFFFFF)), 0, responseBuffer, header.length + 2 + 4, 4);
                    //MNVM_NETIDFILTER_16 @ 8
                    System.arraycopy(BitConverter.getBytes(Integer.MAX_VALUE), 0, responseBuffer, header.length + 2 + 8, 2);
                    //MNVM_SERVERPORT_16 @ 10
                    System.arraycopy(BitConverter.getBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 10, 2);
                    //MNVM_HEARTBEAT_32	@ 12
                    System.arraycopy(BitConverter.getBytes((int)(gateway.ReportInterval * 60)), 0, responseBuffer, header.length + 2 + 12, 4);

                    //No IP Address Settings

                    //MNVM_NetworkListUpdateFrequency_32	@ 32
                    System.arraycopy(BitConverter.getBytes((int)(gateway.NetworkListInterval * 60)), 0, responseBuffer, header.length + 2 + 32, 4);
                    //MNVM_ObserveAware_16	@ 36
                    System.arraycopy(BitConverter.getBytes(BitConverter.toUInt16(gateway.ObserveAware)), 0, responseBuffer, header.length + 2 + 36, 2);

                    //MNVM_ServerNameIP @ 256
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 256, HostAddress.length);
                    //#endregion
                    break;
                case Ethernet_3_0://Ethernet 3.0
                    //#region Ethernet
                    responseBuffer = new byte[header.length + 2 + 512];//512 + 18
                    System.arraycopy(header, 0, responseBuffer, 0, header.length);
                    //Assemble the Configuration page
                    for (int i = 0; i < 512; i++) responseBuffer[header.length + 2 + i] = (byte) 0xFF;  //Base fill with 0xFF
                    HostAddress2 = ParseIP(gateway.ServerHostAddress);


                    Extensions.debug(""+header.length, 31);
                    
                    //APN MANAGEMENT ([4-15])
                    //MNVM_RF_CHAN_32 4 //RF Channel of the APN
                    System.arraycopy(BitConverter.getBytes((int)(0xFFFFFFFF)), 0, responseBuffer, header.length + 2 + 4, 4);
                    //MNVM_RF_NETIDFILTER_16 8 //NET ID of the APN
                    System.arraycopy(BitConverter.getBytes(0xFFFF), 0, responseBuffer, header.length + 2 + 8, 2);

                    Extensions.debug(gateway.GatewayIP, 31);
                    Extensions.debug(gateway.GatewaySubnet, 31);
                    Extensions.debug(gateway.GatewayDNS, 31);
                    Extensions.debug(gateway.SecondaryDNS, 31);
                    Extensions.debug(gateway.DefaultRouterIP, 31);
                    
                    Extensions.debug(Extensions.hexStr(ParseIP(gateway.GatewayIP)), 31);
                    Extensions.debug(Extensions.hexStr(ParseIP(gateway.GatewaySubnet)), 31);
                    Extensions.debug(Extensions.hexStr(ParseIP(gateway.GatewayDNS)), 31);
                    Extensions.debug(Extensions.hexStr(ParseIP(gateway.SecondaryDNS)), 31);
                    Extensions.debug(Extensions.hexStr(ParseIP(gateway.DefaultRouterIP)), 31);
                    
                    //ETHERNET MANAGEMENT ([16-35])
                    //MNVM_MYIP_32 12  //Gateway IP Address (0.0.0.0 denotes DHCP)
                    System.arraycopy(ParseIP(gateway.GatewayIP), 0, responseBuffer, header.length + 2 + 12, 4);
                    //MNVM_MYMASK_32 16  //Gateway IP Mask (0.0.0.0 denotes DHCP)
                    System.arraycopy(ParseIP(gateway.GatewaySubnet), 0, responseBuffer, header.length + 2 + 16, 4);
                    //MNVM_MYDNS1_32 20  //Gateway IP DNS Address 1 (0.0.0.0 denotes DHCP)
                    System.arraycopy(ParseIP(gateway.GatewayDNS), 0, responseBuffer, header.length + 2 + 20, 4);
                    //MNVM_MYDNS2_32 24  //Gateway IP DNS Address 2 (0.0.0.0 denotes DHCP)
                    System.arraycopy(ParseIP(gateway.SecondaryDNS), 0, responseBuffer, header.length + 2 + 24, 4);
                    //MNVM_MYROUTER_32 28  //Gateway IP Address (0.0.0.0 denotes DHCP)
                    System.arraycopy(ParseIP(gateway.DefaultRouterIP), 0, responseBuffer, header.length + 2 + 28, 4);

                    //General Interface Data Management
                    //MNVM_SPBQ_EXP_32 32 //Number seconds that must elapse for automatic DMSG expiration and deletion from Queue.
                    System.arraycopy(BitConverter.getBytes((int)(720 * 60)), 0, responseBuffer, header.length + 2 + 32, 4);
                    //MNVM_MSVR_ACTIVE_8 36 //Is the MSVR Interface Active
                    responseBuffer[header.length + 2 + 36] = BitConverter.ToByte(1);
                    //MNVM_MSGI_ACTIVE_8 37 //Is the MSGI Interface Active
                    responseBuffer[header.length + 2 + 37] = BitConverter.ToByte(0);
                    //MNVM_MODT_ACTIVE_8 38 //Is the ModbusTCP Interface Active
                    responseBuffer[header.length + 2 + 38] = BitConverter.ToByte(0);
                    //MNVM_SNMP_ACTIVE_8 39 //Is the SNMP Interface Active (0x1 - SNMP1 Active, 0x2 - TRAP1 Active, ... 0x40 - SNMP4 Active, 0x80 - Trap4 Active)
                    responseBuffer[header.length + 2 + 39] = 0x00;


                    //MONNIT SERVICE GATEWAY INTERFACE
                    //MNVM_MSGI_TIMEOUT_16 40 //Monnit Service Gateway Interface Timeout in 0.5 seconds(0 means the interface is disabled) 
                    System.arraycopy(BitConverter.getTwoBytes((int)(1 * 60)), 0, responseBuffer, header.length + 2 + 40, 2);
                    //MNVM_MSGI_PORT_16 42 //Monnit Service Gateway Interface Port (3500 default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(3500)), 0, responseBuffer, header.length + 2 + 42, 2);

                    //MODBUS INTERFACE
                    //MNVM_MODBUS_TIMEOUT_16 44 //Modbus TCP Interface Timeout in 0.5 seconds(0 means the interface is disabled) 
                    System.arraycopy(BitConverter.getTwoBytes((int)(5 * 60)), 0, responseBuffer, header.length + 2 + 44, 2);
                    //MNVM_MODBUS_PORT_16 46  //Modbus TCP Interface Port (502 default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(502)), 0, responseBuffer, header.length + 2 + 46, 2);

                    //SNMP INTERFACE
                    //MNVM_SNMP_ADDR1_32 48 //SNMP Socket 1 Address (0L/0xFFFFFFFF means the interface is disabled)
                    //System.arraycopy(ParseIP(gateway.SNMPInterfaceAddress1), 0, responseBuffer, header.length + 2 + 48, 4);
                    //MNVM_SNMP_ADDR2_32 52 //SNMP Socket 2 Address (0L/0xFFFFFFFF means the interface is disabled)
                    //System.arraycopy(ParseIP(gateway.SNMPInterfaceAddress2), 0, responseBuffer, header.length + 2 + 52, 4);
                    //MNVM_SNMP_ADDR3_32 56 //SNMP Socket 3 Address (0L/0xFFFFFFFF means the interface is disabled)
                    //System.arraycopy(ParseIP(gateway.SNMPInterfaceAddress3), 0, responseBuffer, header.length + 2 + 56, 4);
                    //MNVM_SNMP_ADDR4_32 60 //SNMP Socket 4 Address (0L/0xFFFFFFFF means the interface is disabled)
                    //System.arraycopy(ParseIP(gateway.SNMPInterfaceAddress4), 0, responseBuffer, header.length + 2 + 60, 4);
                    //MNVM_SNMP_PORT1_16 64 //SNMP Socket 1 Port (161 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(161)), 0, responseBuffer, header.length + 2 + 64, 2);
                    //MNVM_SNMP_PORT2_16 66 //SNMP Socket 2 Port (161 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(161)), 0, responseBuffer, header.length + 2 + 66, 2);
                    //MNVM_SNMP_PORT3_16 68 //SNMP Socket 3 Port (161 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(161)), 0, responseBuffer, header.length + 2 + 68, 2);
                    //MNVM_SNMP_PORT4_16 70 //SNMP Socket 4 Port (161 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(161)), 0, responseBuffer, header.length + 2 + 70, 2);
                    //MNVM_SNMP_TRAP1_16 72 //SNMP Socket 1 Trap port (162 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(0)), 0, responseBuffer, header.length + 2 + 72, 2);
                    //MNVM_SNMP_TRAP2_16 74 //SNMP Socket 2 Trap port (162 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(0)), 0, responseBuffer, header.length + 2 + 74, 2);
                    //MNVM_SNMP_TRAP3_16 76 //SNMP Socket 3 Trap port (162 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(0)), 0, responseBuffer, header.length + 2 + 76, 2);
                    //MNVM_SNMP_TRAP4_16 78 //SNMP Socket 4 Trap port (162 by default, 0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(0)), 0, responseBuffer, header.length + 2 + 78, 2);

                    //MONNIT SERVER INTERFACE
                    //MNVM_HEARTBEAT_32 80 //Gateway heart beat
                    System.arraycopy(BitConverter.getBytes((int)(gateway.ReportInterval * 60)), 0, responseBuffer, header.length + 2 + 80, 4);
                    //MNVM_GETNETLIST_32 84  //Get network List time
                    System.arraycopy(BitConverter.getBytes((int)(gateway.NetworkListInterval * 60)), 0, responseBuffer, header.length + 2 + 84, 4);
                    //MNVM_DETECTSENALARM_16 88  //Flag to denote that the gateway will respond to urgent messages
                    System.arraycopy(BitConverter.getBytes(BitConverter.toUInt16(gateway.ObserveAware)), 0, responseBuffer, header.length + 2 + 88, 2);
                    //MNVM_SERVERPORT1_16 90  //Server port 1 used with "MNVM_ServerNameIP1"  (0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 90, 2);
                    //MNVM_SERVERPORT2_16 93 //Server port 2 used with "MNVM_ServerNameIP2"  (0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 92, 2);

                    //MNVM_NETPOLL_16 94 //Poll Interval  (0 is disabled)
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.PollInterval * 60)), 0, responseBuffer, header.length + 2 + 94, 2);

                    //MNVM_PROMISCUOUSTIME_16          112    //Promiscuous Time for button behaviors
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway._AutoConfigTime)),0,responseBuffer, header.length+2+112,2);
                    
                    //MNVM_RSV2 96 //IGNORE
                    //MNVM_ServerNameIP1 256 //(Length = 128 or 127 + null)
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 256, HostAddress.length);
                    //MNVM_ServerNameIP2 384 //(Length = 128 or 127 + null)
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 384, HostAddress.length);

                    //#endregion
                    
                    Extensions.debug("GetGatewayConfiguration: " + Extensions.hexStr(responseBuffer), 31);
                    
                    break;
                
                case CGW2://C2GW Verizon
                    //#region C2 Gateway
                    responseBuffer = new byte[header.length + 2 + 512];//512 + 18
                    System.arraycopy(header, 0, responseBuffer, 0, header.length);
                    //Assemble the Configuration page
                    for (int i = 0; i < 512; i++) responseBuffer[header.length + 2 + i] = (byte) 0xFF;  //Base fill with 0xFF


                    //MNVM_RADIOCHANMASK_32 is at Offset 4 (header.length + 4;
                    System.arraycopy(BitConverter.getBytes((int)(0xFFFFFFFF)), 0, responseBuffer, header.length + 2 + 4, 4);
                    //MNVM_NETIDFILTER_16 @ 8
                    System.arraycopy(BitConverter.getTwoBytes(Integer.MAX_VALUE), 0, responseBuffer, header.length + 2 + 8, 2);
                    //MNVM_ObserveAware_16	@ 10
                    System.arraycopy(BitConverter.getTwoBytes(BitConverter.toUInt16(gateway.ObserveAware)), 0, responseBuffer, header.length + 2 + 10, 2);
                    //MNVM_HEARTBEAT_32	@ 12
                    System.arraycopy(BitConverter.getBytes((int)(gateway.ReportInterval * 60)), 0, responseBuffer, header.length + 2 + 12, 4);
                    //MNVM_NetworkListUpdateFrequency_32	@ 16
                    System.arraycopy(BitConverter.getBytes((int)(gateway.NetworkListInterval * 60)), 0, responseBuffer, header.length + 2 + 16, 4);
                    //MNVM_SERVERPORT_16 @ 20
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 20, 2);
                    //MNVM_SERVERPORT_16 @ 22
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 22, 2);

                    //MNVM_NETPOLL_TIMER_16 @ 34
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.PollInterval * 60)), 0, responseBuffer, header.length + 2 + 34, 2);

                    //MNVM_PROMISCUOUSTIME_16          52    //Promiscuous Time for button behaviors
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway._AutoConfigTime)),0,responseBuffer, header.length+2+52,2);
                    
                    //MNVM_ServerNameIP1 192 @ 64
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 64, HostAddress.length);
                    //MNVM_ServerNameIP1 192 @ 256
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 256, HostAddress.length);
                    //#endregion
                    break;
                case CGW3:
                    responseBuffer = new byte[header.length + 2 + 512];//512 + 18
                    System.arraycopy(header, 0, responseBuffer, 0, header.length);
                    //Assemble the Configuration page
                    for (int i = 0; i < 512; i++) responseBuffer[header.length + 2 + i] = (byte) 0xFF;  //Base fill with 0xFF

                    //MNVM_ObserveAware_16	@ 10
                    System.arraycopy(BitConverter.getTwoBytes(BitConverter.toUInt16(gateway.ObserveAware)), 0, responseBuffer, header.length + 2 + 10, 2);
                    //MNVM_HEARTBEAT_32	@ 12
                    System.arraycopy(BitConverter.getBytes((int)(gateway.ReportInterval * 60)), 0, responseBuffer, header.length + 2 + 12, 4);
                    //MNVM_NetworkListUpdateFrequency_32	@ 16
                    System.arraycopy(BitConverter.getBytes((int)(gateway.NetworkListInterval * 60)), 0, responseBuffer, header.length + 2 + 16, 4);
                    //MNVM_SERVERPORT_16 @ 20
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.Port)), 0, responseBuffer, header.length + 2 + 20, 2);

                    //MNVM_NETPOLL_TIMER_16 @ 34
                    System.arraycopy(BitConverter.getTwoBytes((int)(gateway.PollInterval * 60)), 0, responseBuffer, header.length + 2 + 34, 2);
                    //MNVM_SECKEYMASTER_16B //Security Key                              
                    if(gateway.EncryptionKey != null)
                    	System.arraycopy(gateway.EncryptionKey.getEncoded(), 0, responseBuffer, header.length + 2 + 36, 16);
                    else
                    	System.arraycopy(new byte[] {
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF,
                    			(byte) 0xFF
                    	}, 0, responseBuffer, header.length + 2 + 36, 16);

                    //MNVM_DNS1_32 @ 56 
                    byte[] temp = ParseIP(gateway.GatewayDNS);
                    System.arraycopy(temp, 0, responseBuffer, header.length + 2 + 56, 4);
                    //MNVM_DNS2_32 @ 60 
                    temp = ParseIP(gateway.SecondaryDNS);
                    System.arraycopy(temp, 0, responseBuffer, header.length + 2 + 60, 4);
                    //MNVM_APNNAME_64 @ 64 //63+1 
                    temp = NullTermString(gateway.CellAPNName);
                    System.arraycopy(temp, 0, responseBuffer, header.length + 2 + 64, temp.length);
                    //MNVM_USERNAME_64 @ 128 //63+1 
                    temp = NullTermString(gateway.Username);
                    System.arraycopy(temp, 0, responseBuffer, header.length + 2 + 128, temp.length);
                    //MNVM_PASSWORD_64 @ 192 //63+1
                    Password = gateway.Password;
                    temp = NullTermString(Password);
                    System.arraycopy(temp, 0, responseBuffer, header.length + 2 + 192, temp.length);

                    //MNVM_ServerNameIP1 64 @ 384
                    System.arraycopy(HostAddress, 0, responseBuffer, header.length + 2 + 384, HostAddress.length);
                    break;
            }

            //More Flag = 0 (false)
            responseBuffer[15] = 0x00;
            
            Extensions.debug("Exiting GetGatewayConfiguration.");
            return responseBuffer;
        }

        private byte[] RetrieveMessages(Gateway gateway, InboundPacket packet, byte[] header) throws Exception
        {
        	Extensions.debug("Entered RetrieveMessages.", 3);
        	ByteArrayOutputStream ms = new ByteArrayOutputStream();
       //     using (System.IO.MemoryStream ms = new System.IO.MemoryStream())
            
            Boolean More = false;
            int Count = 0;

            int OverallMessageCount = 0;
            int MaxMessageLength = 1024;

            List<Sensor> AffectedSensors = new ArrayList<Sensor>();
            //TODO: Wi-Fi
            //if (gateway.SensorID > 0)//WIFI
            //    AffectedSensors.Add(FindSensor(gateway.SensorID));
            //else
            //{
            Extensions.debug("Reached first for loop in RetrieveMessages.", 3);
            try {
            	if(gateway.AssignedSensors != null) {
            		for(int sid : gateway.AssignedSensors) {
            				Extensions.debug("Cycling first for loop in RetrieveMessages.", 3);
            				AffectedSensors.add(FindSensor(sid));
            		}
            	}
            }
            catch (Exception e) {
            	Extensions.debug("You must register the sensor", 3);
            	OnLogException(e);
            	e.printStackTrace();
            	throw e;
            }
            //}

            
            
            if(gateway.getPendingActionControlCommand()) {
            	MSAPI msg = new MSAPI(Gateway.getAutoConfigCommand(gateway.GatewayID, gateway.AutoConfigCommandTime), MSAPI.OPTIONS_NONE);
            	//System.out.println("adding autoconfigcommand");
            	Count++;
            	ms.write(msg.frame, 0, msg.frame.length);
            }
            else {
            	//System.out.println("not adding autoconfigcommand");
            }
            
            Extensions.debug("Reached second for loop in RetrieveMessages.", 3);    
            for(Sensor s : AffectedSensors)
            {
            	Extensions.debug("Checking sensor config.", 13);
                if (s.getGeneralConfig1Dirty())
                {
                	Extensions.debug("GeneralConfig 1 Dirty.", 13);
                    OverallMessageCount++;

                    if (OverallMessageCount > packet.getMessageCount())
                    {
                        MSAPI msg = new MSAPI(s.ConfigSection(eConfigSection.General_1), MSAPI.OPTIONS_NONE);
                        if (ms.size() + msg.frame.length > MaxMessageLength) {
                        	Extensions.debug("Message size too small.", 13);
                        	break;
                        }


                        Count++;
                        Extensions.debug("Writing message in GeneralConfig1", 14);
                        ms.write(msg.frame, 0, msg.frame.length);
                       // s.setGeneralConfig1Dirty(false);
                        Extensions.debug("Wrote message in GeneralConfig1", 14);
                    }
                }
                if (s.getGeneralConfig2Dirty())
                {
                	Extensions.debug("GeneralConfig2Dirty.", 12);
                    OverallMessageCount++;

                    if (OverallMessageCount > packet.getMessageCount())
                    {
                        MSAPI msg = new MSAPI(s.ConfigSection(eConfigSection.General_2), MSAPI.OPTIONS_NONE);
                        if (ms.size() + msg.frame.length > MaxMessageLength)
                            break;


                        Count++;
                        ms.write(msg.frame, 0, msg.frame.length);
                        //s.setGeneralConfig2Dirty(false);
                    }
                }
                
                Extensions.debug("Halfway through sensor config.", 12);
                try {
                if (s.getProfileConfig1Dirty())
                {
                	Extensions.debug("ProfileConfig1Dirty.", 12);
                    OverallMessageCount++;

                    if (OverallMessageCount > packet.getMessageCount())
                    {
                        MSAPI msg = new MSAPI(s.ConfigSection(eConfigSection.Profile_1), MSAPI.OPTIONS_NONE);
                        if (ms.size() + msg.frame.length > MaxMessageLength) {
                            Extensions.debug("Breaking for messagesize restrictions.", 12);
                        	break;
                        }


                        Count++;
                        Extensions.debug("Writing message. ProfileConfig", 12);
                        ms.write(msg.frame, 0, msg.frame.length);
                    }
                    else {
                    	Extensions.debug("Profile1 skipped if statement", 12);
                    }
                }
                }
                catch (Exception e) {
                	e.printStackTrace();
                	OnLogException(e);
                }
                
                Extensions.debug("Almost done with sensor config.", 12);
                
                if (s.getProfileConfig2Dirty())
                {
                	Extensions.debug("ProfileConfig2Dirty.", 12);
                    OverallMessageCount++;

                    if (OverallMessageCount > packet.getMessageCount())
                    {
                        MSAPI msg = new MSAPI(s.ConfigSection(eConfigSection.Profile_2), MSAPI.OPTIONS_NONE);
                        if (ms.size() + msg.frame.length > MaxMessageLength)
                            break;


                        Count++;
                        ms.write(msg.frame, 0, msg.frame.length);
                    }
                }
                
                Extensions.debug("3/4s through checking sensor config.", 12);
                try {
                
                if (s.getPendingActionControlCommand() && s.getActionControlCommands() != null)
                {
                    Boolean isFirst = true;
                    List<Integer> successfulIndeces = new ArrayList<Integer>();
                    for(byte[] command : s.getActionControlCommands())
                    {
                        OverallMessageCount++;

                        if (OverallMessageCount > packet.getMessageCount())
                        {
                            MSAPI msg = new MSAPI(command, (isFirst && s.getActionControlCommandIsUrgent()) ? MSAPI.OPTIONS_URGENT : MSAPI.OPTIONS_NONE);
                            if (ms.size() + msg.frame.length > MaxMessageLength)
                                break;

                            isFirst = false;

                            Count++;
                            ms.write(msg.frame, 0, msg.frame.length);

                            //Local Notifier could have many messages so check against max length and break if there are too many
                            if (Count > 290)
                                break;
                            
                            
//                            System.out.println("Finished 'if' in PendingCommands.");
                        }
//                        System.out.println("\n\nProcessed command.\n\n");
                    }
                }
                }
                catch (Exception e) {
                	e.printStackTrace();
                	OnLogException(e);
                	throw e;
                }

                Extensions.debug("if count > 290 checking sensor config.", 12);
                
                if (Count > 290)//300 max for gateway 290 there are 8 possible down messages per sensor so 290 ensures we are always under 300
                {
                    More = true;
                    break;
                }
            }

            Extensions.debug("Finished checking sensor config.", 13);
            
            byte[] Messages = ms.toByteArray();
            

            byte[] responseBuffer = new byte[header.length + 2 + Messages.length];
            System.arraycopy(header, 0, responseBuffer, 0, header.length);
            
            //Count of Pending Messages
            Extensions.debug(""+header.length, 14);
            System.arraycopy(BitConverter.getBytes(Count), 0, responseBuffer, header.length, 2);
            
            //Messages
            System.arraycopy(Messages, 0, responseBuffer, header.length + 2, Messages.length);

            //More Flag
            responseBuffer[15] = BitConverter.toByte(More);
            
            StringBuffer sb = new StringBuffer();
            for(byte b : responseBuffer) {
            	sb.append(String.format("%02X-", b));
            }
          //  System.out.println(sb.toString());

            Extensions.debug("Leaving RetrieveMessages.", 14);
            return responseBuffer;        
        }
        //#endregion

        //#region Command 2 SensorIDList
        private byte[] SensorIDList(Gateway gateway, InboundPacket packet, byte[] header) throws Exception
        {
        	Extensions.debug("Entered SensorIDList", 30);
        	try {
	            Boolean More = false;
	            int OverallSensorCount = 0;
	            List<Sensor> list = new ArrayList<Sensor>();
	
	            //Order by SensorID
	            for(int sid : gateway.getAssignedSensors())
	            {
	                OverallSensorCount++;
	
	                // skip sensors already sent in previous batch with more flag set
	                if (OverallSensorCount <= packet.getMessageCount())
	                    continue;
	                
	                Sensor s = FindSensor(sid);
	                
	                list.add(s);
	
	                if (list.size() > 199)//Only allow 200 sensors per network  Or address with More flag
	                {
	                    More = true;
	                    break;
	                }
	            }
	
	            byte[] responseBuffer = new byte[header.length + 2 + (list.size() * 4)];
	            System.arraycopy(header, 0, responseBuffer, 0, header.length);
	            //Count of SensorIDs
	            System.arraycopy(BitConverter.getBytes((int)(list.size())), 0, responseBuffer, header.length, 2);
	            //SIDs
	            int CurrentIndex = header.length + 2;
	            for(Sensor s : list)
	            {
	                System.arraycopy(BitConverter.getBytes((int)(s.SensorID)), 0, responseBuffer, CurrentIndex, 4);
	                CurrentIndex = CurrentIndex + 4;
	            }
	
	            if (gateway.RadioFirmwareVersion.equals("2.2.0.0"))
	            {
	                //Set More Flag will only be set if we allow more than 200 sensors to be sent to an apn
	                responseBuffer[15] = BitConverter.toByte(More);
	            }
	            
	            Extensions.debug("Exited SensorIDList", 30);
	            return responseBuffer;
        	}
        	catch (Exception e) {
        		e.printStackTrace();
        		OnLogException(e);
        		throw e;
        	}
        }

        private byte[] SaveGatewayConfiguration(Gateway gateway, byte[] header, byte[] configurationPage, int commVersion) throws Exception
        {
        	
        	Extensions.debug("Entered SaveGatewayConfiguration", 8);
        	try {
        		//byte[] tempByte = new byte[4];
        		//System.arraycopy(configurationPage,4, tempByte, 0, 4);
            int DHL =16; //BitConverter.ToInt32(tempByte); // Decryption Helper Length
            switch (gateway.GatewayType)
            {
                case USB://USB Service Gateway
                	Extensions.debug("USB Case", 8);
                    //#region USB Service
                    //MNVM_NETIDFILTER_16 @ 8
                    //gateway.NetworkIDFilter = BitConverter.ToUInt16(configurationPage, DHL + 8);
                    //MNVM_SERVERPORT_16 @ 10
                	
                	//gateway.setPort(BitConverter.ToUInt16(configurationPage, DHL + 10));
                	gateway.setPort(BitConverter.twoBytesToInt(configurationPage, DHL + 10));
                	//gateway.setPort((int)configurationPage[DHL + 10] & 0xFF);
                	Extensions.debug("New Port " + gateway.getPort(), 12);
                	
                    //MNVM_HEARTBEAT_32	@ 12
                	
                    gateway.ReportInterval = (double)(BitConverter.toInt32(configurationPage, DHL + 12)) / 60.0;
                    Extensions.debug("ReportInterval " + gateway.getReportInterval(), 8);


                    //MNVM_NetworkListUpdateFrequency_32	@ 32
                    gateway.NetworkListInterval = (int)(BitConverter.toInt32(configurationPage, DHL + 32) / 60);
                    //MNVM_ObserveAware_16	@ 36
                    gateway.ObserveAware = BitConverter.ToUInt16(configurationPage, DHL + 36) != 0;

                    //MNVM_ServerNameIP @ 256
                    if (configurationPage[16 + 256 + 4] == 0x00) {
                    	Extensions.debug("Entered if.", 8);
                    	gateway.ServerHostAddress = ExtractIP(configurationPage, DHL + 256);
                    }
                    else {
                    	Extensions.debug("Entered Else", 8);
                        gateway.ServerHostAddress = ExtractNullTermString(configurationPage, DHL + 256, 200);
                    }
                    //#endregion
                    break;
                case Ethernet_3_0://Ethernet 3.0
                    //#region Ethernet 3.0

                	Extensions.debug("Getting eth gate info", 30);
                    //ETHERNET MANAGEMENT ([16-35])
                    //MNVM_MYIP_32 12  //Gateway IP Address (0.0.0.0 denotes DHCP)
                    gateway.GatewayIP = ExtractIP(configurationPage, DHL + 12);
                    //MNVM_MYMASK_32 16  //Gateway IP Mask (0.0.0.0 denotes DHCP)
                    gateway.GatewaySubnet = ExtractIP(configurationPage, DHL + 16);
                    //MNVM_MYDNS1_32 20  //Gateway IP DNS Address 1 (0.0.0.0 denotes DHCP)
                    gateway.GatewayDNS = ExtractIP(configurationPage, DHL + 20);
                    //MNVM_MYDNS2_32 24  //Gateway IP DNS Address 2 (0.0.0.0 denotes DHCP)
                    gateway.SecondaryDNS = ExtractIP(configurationPage, DHL + 24);
                    //MNVM_MYROUTER_32 28  //Gateway IP Address (0.0.0.0 denotes DHCP)
                    gateway.DefaultRouterIP = ExtractIP(configurationPage, DHL + 28);

                    Extensions.debug(gateway.GatewayIP, 30);
                    Extensions.debug(gateway.GatewaySubnet, 30);
                    Extensions.debug(gateway.GatewayDNS, 30);
                    Extensions.debug(gateway.SecondaryDNS, 30);
                    Extensions.debug(gateway.DefaultRouterIP, 30);
                    
                    //MONNIT SERVER INTERFACE
                    //MNVM_HEARTBEAT_32 80 //Gateway heart beat
                    gateway.ReportInterval = (double)(BitConverter.toInt32(configurationPage, DHL + 80)) / 60;
                    //MNVM_GETNETLIST_32 84  //Get network List time
                    gateway.NetworkListInterval = (int)(BitConverter.toInt32(configurationPage, DHL + 84) / 60);
                    //MNVM_DETECTSENALARM_16 88  //Flag to denote that the gateway will respond to urgent messages
                    gateway.ObserveAware = BitConverter.ToBoolean(BitConverter.ToUInt16(configurationPage, DHL + 88));
                    //MNVM_SERVERPORT1_16 90  //Server port 1 used with "MNVM_ServerNameIP1"  (0 is disabled)
                    gateway.Port = BitConverter.twoBytesToInt(configurationPage, DHL + 90);
                    //MNVM_NETPOLL_16 94 //Poll Interval  (0 is disabled)
                    gateway.PollInterval = BitConverter.twoBytesToInt(configurationPage, DHL + 94) / 60d;
                    
                    //MNVM_PROMISCUOUSTIME_16          112    //Promiscuous Time for button behaviors
                    gateway._AutoConfigTime = BitConverter.twoBytesToInt(configurationPage, DHL+112);
                    
                    //MNVM_RSV2_16 96 //IGNORE
                    //MNVM_ServerNameIP1 256 //(Length = 128 or 127 + null)
                    if (configurationPage[DHL + 256 + 4] == 0x00)
                        gateway.ServerHostAddress = ExtractIP(configurationPage, DHL + 256);
                    else
                        gateway.ServerHostAddress = ExtractNullTermString(configurationPage, DHL + 256, 128);
                    
                    Extensions.debug("Done getting eth gate info", 30);
                    //#endregion
                    break;
                case CGW2://C2GW Dev
                    //case 22://C3GW Verizon
                    //case 23://C3GW Sprint
                    //#region C2GW
                    //MNVM_ObserveAware_16	@ 10
                    gateway.ObserveAware = BitConverter.ToBoolean(BitConverter.ToUInt16(configurationPage, DHL + 10));
                    //MNVM_HEARTBEAT_32	@ 12
                    gateway.ReportInterval = BitConverter.ToDouble(BitConverter.toInt32(configurationPage, DHL + 12)) / 60;
                    //MNVM_NetworkListUpdateFrequency_32	@ 16
                    gateway.NetworkListInterval = (int)(BitConverter.toInt32(configurationPage, DHL + 16) / 60);
                    //MNVM_SERVERPORT_16 @ 20
                    gateway.Port = BitConverter.twoBytesToInt(configurationPage, DHL + 20);
                    //MNVM_PollInterval_32	@ 34
                    gateway.PollInterval = BitConverter.toInt32(configurationPage, DHL + 34) / 60d;

                    //MNVM_PROMISCUOUSTIME_16          52    //Promiscuous Time for button behaviors
                    gateway._AutoConfigTime = BitConverter.twoBytesToInt(configurationPage, DHL+52);
                    
                    //MNVM_ServerNameIP 192 @ 64
                    if (configurationPage[16 + 64 + 4] == 0x00)
                        gateway.ServerHostAddress = ExtractIP(configurationPage, DHL + 64);
                    else
                        gateway.ServerHostAddress = ExtractNullTermString(configurationPage, DHL + 64, 200);
                    //#endregion
                    break;
                case CGW3:
                    //MNVM_ObserveAware_16	@ 10
                    gateway.ObserveAware = BitConverter.ToBoolean(BitConverter.twoBytesToInt(configurationPage, DHL + 10));
                    //MNVM_HEARTBEAT_32	@ 12
                    gateway.ReportInterval = BitConverter.ToDouble(BitConverter.ToUInt32(configurationPage, DHL + 12)) / 60d;
                    //MNVM_NetworkListUpdateFrequency_32	@ 16
                    gateway.NetworkListInterval = BitConverter.ToDouble(BitConverter.ToUInt32(configurationPage, DHL + 16)) / 60d;
                    //MNVM_SERVERPORT_16 @ 20
                    gateway.Port = BitConverter.twoBytesToInt(configurationPage, DHL + 20);
                    //MNVM_PollTimmer_16 @ 34
                    gateway.PollInterval = BitConverter.ToDouble(BitConverter.twoBytesToInt(configurationPage, DHL + 34)) / 60d;

                    //MNVM_SECKEYMASTER_16B //Security Key
                    byte[] newkey = new byte[16];
                    System.arraycopy(configurationPage, DHL + 36, newkey, 0, 16);
                    gateway.EncryptionKey = new SecretKeySpec(newkey, "AES");

                    gateway._AutoConfigTime = BitConverter.twoBytesToInt(configurationPage, DHL + 52);

                    //MNVM_DNS1_32 @ 56 
                    gateway.GatewayDNS = ExtractIP(configurationPage, DHL + 56);
                    //MNVM_DNS2_32 @ 60
                    gateway.SecondaryDNS = ExtractIP(configurationPage, DHL + 60);
                    //MNVM_APNNAME_64 @ 64 //63+1 
                    gateway.CellAPNName = ExtractNullTermString(configurationPage, DHL + 64, 64);
                    //MNVM_USERNAME_64 @ 128 //63+1 
                    gateway.Username = ExtractNullTermString(configurationPage, DHL + 128, 64);
                    //MNVM_PASSWORD_64 @ 192 //63+1
                    gateway.Password = ExtractNullTermString(configurationPage, DHL + 192, 64);


                    //MNVM_ServerNameIP 64 @ 384
                    if (configurationPage[16 + 384 + 4] == 0x00)
                        gateway.ServerHostAddress = ExtractIP(configurationPage, DHL + 384);
                    else
                        gateway.ServerHostAddress = ExtractNullTermString(configurationPage, DHL + 384, 64);
                    break;                    
            }
        	}
        	catch(Exception e) {
        		Extensions.debug("Caught exception while getting eth gate info", 30);
        		Extensions.debug(e.toString(), 30);
        		OnLogException(e);
        		e.printStackTrace();
        		throw e;
        	}
            gateway.IsDirty = false;

            Extensions.debug("Exited SaveGatewayConfiguration with Empty Ack.", 30);
            return ReturnEmptyAck(header);
        }

        //#region Command 11 Gateway Startup
        private void RecordGatewayStartup(Gateway sensorGateway, InboundPacket packet) throws Exception
        {
        	Extensions.debug("Entered RecordGatewayStartup", 11);
        	try {
            String FirmwareVersion = (int)(packet.getPayload()[0]) + "." + (int)(packet.getPayload()[1]) + "." + (int)(packet.getPayload()[2]) + "." + (int)(packet.getPayload()[3]);
            sensorGateway.GatewayFirmwareVersion = FirmwareVersion;

            switch (sensorGateway.GatewayType)
            {
                case USB:
                	Extensions.debug("USB Case just goes straight to break", 11);
                    break;
                case CGW2:
                    long MEID = (long) BitConverter.ToDouble(packet.getPayload(), 4);
                    String Identifiers = ""+ MEID + "|";
                    
                    Extensions.debug("Packet Payload: " + Extensions.hexStr(packet.getPayload()), 35);
                    
                    for (int i = 12; i < 22; i++)
                        Identifiers += (char)packet.getPayload()[i];

                    sensorGateway.ExternalIdentifier = Identifiers;
                    break;
                case Ethernet_3_0:
                    //if (packet.getMessageCount() == 7 && sensorGateway.GatewayType != eGatewayType.Ethernet_3_0)//EGW 3.0 firmware but wasn't set as type 7
                    //{
                    //    sensorGateway.GatewayType = eGatewayType.Ethernet_3_0;
                    //}
                    //OnLogException(new Exception("Incorrect Gateway Type"));
                    try {
                    	byte[] EGWMacBytes = new byte[6];
                    	System.arraycopy(packet.getPayload(), 4, EGWMacBytes, 0, 6);
                    	String EGWMacAddress = BitConverter.ToString(EGWMacBytes).replace("-", "");
                    	sensorGateway.ExternalIdentifier = EGWMacAddress;
                    }
                    catch (Exception e) {
                    	Extensions.debug("RecordGatewayStartup Failed.", 30);
                    }

                    break;
                case CGW3:
                    MEID = (long) BitConverter.ToDouble(packet.getPayload(), 4);
                    Identifiers = "" + MEID + "|";
                    for (int i = 12; i < 22; i++)
                        Identifiers += (char)packet.getPayload()[i];

                    sensorGateway.ExternalIdentifier = Identifiers;
                    break;
                default:
                    byte[] MacBytes = new byte[6];
                    System.arraycopy(packet.Payload, 4, MacBytes, 0, 6);
                    String MacAddress = BitConverter.ToString(MacBytes).replace("-", "");
                    if (packet.getMessageCount() == 4)//4 is Ethernet Gateway (only V2 sends this message)
                    {
                        sensorGateway.ExternalIdentifier = MacAddress;
                    }
                    break;
            }
        	}
        	catch (Exception e) {
        		OnLogException(e);
        		e.printStackTrace();
        		throw e;
        	}
            Extensions.debug("Exited RecordGatewayStartup", 11);
        }

        byte[] ReturnStartupPacket(Gateway gateway, byte[] header) throws Exception
        {
        	Extensions.debug("Entered ReturnStartupPacket", 50);
            byte[] responseBuffer = new byte[header.length + 2 + 48];
            
            
            //If GatewayKey is valid apply security settings
            if(IsSecurityKeyValid(gateway.PreSharedKey))
            {
            	Extensions.debug("Entering Security check", 50);
                //Create new SecurityKey and IV for this session with gateway
                gateway.EncryptionKey = GW_Encryption.KeyGenerator();
                Extensions.debug("\n\n\nPermanent Key: " + Extensions.hexStr(gateway.EncryptionKey.getEncoded()), 30);
                gateway.EncryptionIV = GW_Encryption.IVGenerator();
                Extensions.debug("Permanent IV: " + Extensions.hexStr(gateway.EncryptionIV), 30);
                byte[] IV = GW_Encryption.IVGenerator();

                //Create Payload and encrypt
                EncResult Payload = new EncResult();
                Payload.result = new byte[32];
                System.arraycopy(gateway.EncryptionKey.getEncoded(), 0, Payload.result, 0, 16);
                System.arraycopy(gateway.EncryptionIV, 0, Payload.result, 16, 12);
                System.arraycopy(gateway.PreSharedKey.getEncoded(), 12, Payload.result, 28, 4);

                
                Payload = GW_Encryption.Encrypt128(Payload.result, gateway.PreSharedKey, IV);

                System.arraycopy(Payload.result, 0, responseBuffer, header.length + 2, 32);
                System.arraycopy(IV, 0, responseBuffer, header.length + 2 + 32, 16);
            }
            else
            {
            	Extensions.debug("Key not valid", 50);
                responseBuffer = new byte[header.length + 2];
            }
            
            System.arraycopy(header, 0, responseBuffer, 0, header.length);
            //Count of Messages
            System.arraycopy(BitConverter.getBytes(0), 0, responseBuffer, header.length, 2);

            Extensions.debug("Exited ReturnStartupPacket");
            return responseBuffer;
        }
        
        private boolean IsSecurityKeyValid(SecretKey securityKey)
        {
        	if(securityKey == null) return false;
        	byte[] byteKey = securityKey.getEncoded();
            if (byteKey == null)
                return false; //Not Valid

            try
            {
                //All 00
                boolean AllZero = true;
                for (int i = 0; i < 16; i++)
                {
                    if (byteKey[i] != 0x00)
                    {
                        AllZero = false;
                        break;
                    }
                }
                if (AllZero)
                    return false; //Not Valid


                //All FF
                boolean AllMaxVal = true;
                for (int i = 0; i < 16; i++)
                {
                    if (byteKey[i] != 0xFF)
                    {
                        AllMaxVal = false;
                        break;
                    }
                }
                if (AllMaxVal)
                    return false; //Not Valid

                //Not Null
                //Not all 00
                //Not all FF
                return true; //Valid
            }
            catch (Exception e)
            {
                return false; //Not Valid
            }
        }

        //#endregion
        //#endregion

        //#region Common Helper Methods
        byte[] ResponseHeader(InboundPacket packet)
        {
            try
            {
                byte[] header = new byte[16];
                if (packet.getVersion() == 1)
                {
                    //CSNET
                    System.arraycopy(BitConverter.getBytes((int)(0)), 0, header, 0, 4);
                    //APNID
                    System.arraycopy(BitConverter.getBytes((int)(packet.getGatewayID())), 0, header, 4, 4);
                }
                else
                {
                    //CSNET
                    System.arraycopy(BitConverter.getBytes((int)(packet.getGatewayID())), 0, header, 0, 4);
                    //Security
                    System.arraycopy(BitConverter.getBytes((int)(packet.getSecurity())), 0, header, 4, 4);
                }
                //Version
                header[8] = (byte)(packet.getVersion());

                //Sequence
                header[9] = (byte)(packet.getSequence());

                //Time (v2 UTC seconds since Jan 1 2010)
                long time =  SecondsSinceFirstDate();
                byte[] timeBytes = BitConverter.getFourBytes(time);
                System.arraycopy(timeBytes, 0, header, 10, 4);
                
                header[14] = BitConverter.ToByte(packet.getResponse());

                return header;
            }
            catch (Exception ex)
            {
                OnLogException(ex, "MineServer.ResponseHeader(InboundPacket packet)");
                return null;
            }
        }

        /**
         * <p>
         * gets how many seconds its been since January 1st 2010 to set date times<br/>
         * returns the time based off of UTC
         * </p>/
         */
        public static long SecondsSinceFirstDate()
        { 
        	Calendar now = Calendar.getInstance(TimeZone.getTimeZone("UTC"));

    		Calendar startDate = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
    		startDate.set(2010, 0, 1, 0, 0, 0);

    		long nowSeconds = now.getTimeInMillis() - startDate.getTimeInMillis();
    		long secs = nowSeconds / 1000;

    		return secs;
        }
        
        /**
         * <p>
         * Returns a UTC Calendar based on the seconds passed in 
         * </p>
         * 
         * @param seconds the seconds since January 1st 2010 
         */
        public static Calendar GetCalendarFromSeconds(long seconds)
        {
        	Calendar startDate = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        	startDate.set(2010, 0, 1, 0, 0, 0);
        		
        	//long millis = (seconds * 1000);
        	//long checkin = startDate.getTimeInMillis() + millis;
        	//startDate.setTimeInMillis(checkin);
        	
        	startDate.add(Calendar.SECOND, (int)seconds);
        	return startDate;
        }
        
        /**
         * <p>
         * Returns a UTC Calendar based on the seconds passed in 
         * </p>
         * 
         * @param seconds the seconds since January 1st 2010 
         */
        public static Calendar CurrentUTC()
        {
        	return Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        }
        
        /**
         * <p>
         * Returns a UTC Calendar object
         * </p>
         * 
         * @param calendar Calendar object used for date and time information
         */
        public static Calendar UpdateTimeZoneToUTC(Calendar calendar)
        {
        	Calendar startDate = Calendar.getInstance(TimeZone.getTimeZone("UTC"));
        	
        	startDate.setTime(calendar.getTime());
        	
        	return startDate;
        }
        
        /**
         * <p>
         * Returns a simpleDateFormat of ("E MMM dd yyyy hh:mm:ss a")<br />
         * date and time will look like -> Fri Jan 01 2010 12:00:00 PM. <br />
         * for display purposes
         * </p>
         * 
         * @param calendar Calendar object used for date and time information
         */
       public static String DisplayDateAsUTC(Calendar calendar)
       {
    	   SimpleDateFormat  sdf = new SimpleDateFormat("E MMM dd yyyy hh:mm:ss a");
    	      sdf.setTimeZone(TimeZone.getTimeZone("UTC"));
    	   
    	   return sdf.format(calendar.getTime());
       }
        
        
        private byte[] UpdateResponse(byte[] responseBuffer, int responseCommand)
        {
            try
            {
            	responseBuffer[14] = (byte) responseCommand;
                //System.arraycopy(BitConverter.getBytes(responseCommand), 0, responseBuffer, 14, 1);
                //return responseBuffer;
            }
            catch (Exception ex)
            {
            	ex.printStackTrace();
                OnLogException(ex, "MineServer.UpdateResponse(byte[] responseBuffer, int responseCommand)");
            }
            return responseBuffer;
        }

        private byte[] ReturnZeroAck(byte[] header) throws InterruptedException
        {
        	Extensions.debug("Entered ReturnZeroAck");
            byte[] responseBuffer = new byte[header.length + 2];
            System.arraycopy(header, 0, responseBuffer, 0, header.length);
            //Count of Messages
            System.arraycopy(BitConverter.getBytes(0), 0, responseBuffer, header.length, 2);

            //change response type to 0 for ack
            responseBuffer[14] = (byte)(0);
        	Extensions.debug("Exited ReturnZeroAck");
            return responseBuffer;
        }

        private byte[] ReturnEmptyAck(byte[] header)
        {
            byte[] responseBuffer = new byte[header.length + 2];
            System.arraycopy(header, 0, responseBuffer, 0, header.length);

            return responseBuffer;
        }

        private void FillVariableLengthConfig(byte[] configurationPage, int start, int maxLength, byte[] value)
        {
            byte[] temp2 = new byte[maxLength];
            System.arraycopy(temp2, 0, configurationPage, start, maxLength);
            System.arraycopy(value, 0, configurationPage, start, value.length > maxLength ? maxLength : value.length);

        }

        private String ExtractIP(byte[] bytes, int startIndex)
        {
            return ((int)bytes[startIndex] & 0xFF) + "." + ((int)bytes[startIndex + 1] & 0xFF) + "." + ((int)bytes[startIndex + 2] & 0xFF) + "." + ((int)bytes[startIndex + 3] & 0xFF);
        }

        public byte[] ParseIP(String hostAddress)
        {
            Boolean isIP = false;
            byte[] HostAddress;
            String[] ipSections = hostAddress.split("\\.");
            if (ipSections.length == 4)
            {
                isIP = true;//Maybe
                for(String section : ipSections)
                {
                	try {
                		int Temp = Integer.parseInt(section);
                		Extensions.debug(section, 31);
                		if(Temp < 0 || Temp > 255) {
                			isIP = false;
                		}
                	}
                    catch (Exception e) {
                    	isIP=false;
                    	break;
                    }
                }
            }
            if (isIP)
            {
                int Temp;

                HostAddress = new byte[5];
                HostAddress[0] = BitConverter.ToByte(Integer.parseInt(ipSections[0]));
                HostAddress[1] = BitConverter.ToByte(Integer.parseInt(ipSections[1]));
                HostAddress[2] = BitConverter.ToByte(Integer.parseInt(ipSections[2]));
                HostAddress[3] = BitConverter.ToByte(Integer.parseInt(ipSections[3]));
                HostAddress[4] = BitConverter.ToByte(0);
            }
            else if (Extensions.stringIsNullOrEmpty(hostAddress))
            {
                HostAddress = new byte[5];
                HostAddress[0] = (byte) 0xFF;
                HostAddress[1] = (byte) 0xFF;
                HostAddress[2] = (byte) 0xFF;
                HostAddress[3] = (byte) 0xFF;
                HostAddress[4] = BitConverter.ToByte(0);
            }
            else
            {
                HostAddress = NullTermString(hostAddress);
            }

            return HostAddress;
        }

        private byte[] NullTermString(String str)
        {
        	if(str == null) str = "";
            byte[] temp = str.getBytes(Charset.forName("ASCII"));
            byte[] retval = new byte[temp.length + 1];
            System.arraycopy(temp, 0, retval, 0, temp.length);
            //last byte left 0 for null terminator of String

            return retval;
        }

        private String ExtractNullTermString(byte[] bytes, int startIndex, int length) throws InterruptedException
        {
            StringBuilder sb = new StringBuilder();
            Charset cs = Charset.forName("ASCII");
           	String s = new String(bytes, cs);
           	//String s = new String(bytes, startIndex, startIndex+length, cs);
             for(char c : s.toCharArray())
            {
                if (c == '\0')
                    break;
                sb.append(c);
                Extensions.debug("Cycling ExtractNullTermString", 8);
            }

            return sb.toString();
        }
        
        
        ////////
        //JAVA Event Handling
        ///////
        
        private void HandleGatewayMessageProcessing(GatewayMessage gmsg) throws InterruptedException {
        	Extensions.debug("Reached HandleDataProcessing with "+ GatewayProcessingHandlers.size() + " Handlers", 0);
        	for(iGatewayMessageHandler handler : GatewayProcessingHandlers) {
        		new GatewayMessageHandlerThread(handler, gmsg).start();
        		Extensions.debug("Cycling HandleDataProcessing.", 0);
        		//handler.ProcessGatewayMessage(gmsg);
        	}
        }
        
        private void HandleSensorMessageProcessing(List<SensorMessage> sensorMessageList, Gateway gate) {
        	//System.out.println("Running SensorMessageHanderProcessing");
        	for(iSensorMessageHandler handler : SensorProcessingHandlers) {
        		//System.out.println("Cycling SensorMessageHanderThread");
        		new SensorMessageHandlerThread(handler, sensorMessageList, gate).start();
        		//handler.ProcessSensorMessages(sensorMessageList, gate);
        	}
        }
        
        private void HandleExceptionProcessing(Exception e) {
        	for(iExceptionHandler handler : ExceptionHandlers) {
        		new ExceptionHandlerThread(handler, e).start();
        		//handler.ProcessSensorMessages(sensorMessageList, gate);
        	}
        }
        
        private void HandlePersistSensorProcessing(long sid) {
        	for(iPersistSensorHandler handler : PersistSensorHandlers) {
        		new PersistSensorHandlerThread(handler, sid).start();
        		//handler.ProcessSensorMessages(sensorMessageList, gate);
        	}
        }
        
        private void HandlePersistGatewayProcessing(long gid) {
        	for(iPersistGatewayHandler handler : PersistGatewayHandlers) {
        		new PersistGatewayHandlerThread(handler, gid).start();
        		//handler.ProcessSensorMessages(sensorMessageList, gate);
        	}
        }
        
        private List<iSensorMessageHandler> SensorProcessingHandlers = new ArrayList<iSensorMessageHandler>();
        public boolean addSensorDataProcessingHandler(iSensorMessageHandler handler) throws InterruptedException {
        	try {
        		Extensions.debug("There were " + SensorProcessingHandlers.size() + " handlers" , 0);
        		SensorProcessingHandlers.add(handler);
        		Extensions.debug("There are now " + SensorProcessingHandlers.size() + " handlers" , 0);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString(), 0);
        		return false;
        	}
        }
        public boolean removeSensorDataProcessingHandler(iSensorMessageHandler handler) throws InterruptedException {
        	try {
        		SensorProcessingHandlers.remove(handler);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString());
        		return false;
        	}
        }
        
        private List<iGatewayMessageHandler> GatewayProcessingHandlers = new ArrayList<iGatewayMessageHandler>();
        public boolean addGatewayDataProcessingHandler(iGatewayMessageHandler handler) throws InterruptedException {
        	try {
        		Extensions.debug("There were " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		GatewayProcessingHandlers.add(handler);
        		Extensions.debug("There are now " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString(), 0);
        		return false;
        	}
        }
        public boolean removeGatewayDataProcessingHandler(iGatewayMessageHandler handler) throws InterruptedException {
        	try {
        		GatewayProcessingHandlers.remove(handler);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString());
        		return false;
        	}
        }
        
        private List<iExceptionHandler> ExceptionHandlers = new ArrayList<iExceptionHandler>();
        public boolean addExceptionProcessingHandler(iExceptionHandler handler) throws InterruptedException {
        	try {
        		Extensions.debug("There were " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		ExceptionHandlers.add(handler);
        		Extensions.debug("There are now " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString(), 0);
        		return false;
        	}
        }
        public boolean removeExceptionProcessingHandler(iExceptionHandler handler) throws InterruptedException {
        	try {
        		ExceptionHandlers.remove(handler);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString());
        		return false;
        	}
        }
        
        private List<iPersistGatewayHandler> PersistGatewayHandlers = new ArrayList<iPersistGatewayHandler>();
        public boolean addPersistGatewayHandler(iPersistGatewayHandler handler) throws InterruptedException {
        	try {
        		Extensions.debug("There were " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		PersistGatewayHandlers.add(handler);
        		Extensions.debug("There are now " + GatewayProcessingHandlers.size() + " handlers" , 0);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString(), 0);
        		return false;
        	}
        }
        public boolean removePersistGatewayHandler(iPersistGatewayHandler handler) throws InterruptedException {
        	try {
        		PersistGatewayHandlers.remove(handler);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString());
        		return false;
        	}
        }
        
        private List<iPersistSensorHandler> PersistSensorHandlers = new ArrayList<iPersistSensorHandler>();
        public boolean addPersistSensorHandler(iPersistSensorHandler handler) throws InterruptedException {
        	try {
        		Extensions.debug("There were " + SensorProcessingHandlers.size() + " handlers" , 0);
        		PersistSensorHandlers.add(handler);
        		Extensions.debug("There are now " + SensorProcessingHandlers.size() + " handlers" , 0);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString(), 0);
        		return false;
        	}
        }
        public boolean removePersistSensorHandler(iPersistSensorHandler handler) throws InterruptedException {
        	try {
        		PersistSensorHandlers.remove(handler);
        		return true;
        	}
        	catch (Exception e) {
        		Extensions.debug(e.toString());
        		return false;
        	}
        }
        
        ///////
        //      GATEWAY RESPONSE HANDLING
        ///////
        
        public List<MineServiceGatewayResponseAccess> GatewayResponseHandlers = new ArrayList<MineServiceGatewayResponseAccess>();
        public void addGatewayResponseHandler(MineServiceGatewayResponseAccess handler) {
        	GatewayResponseHandlers.add(handler);
        }
        public boolean removeGatewayResponseHandler(MineServiceGatewayResponseAccess handler) {
        	return GatewayResponseHandlers.remove(handler);
        }
   
        public void OnNetworkStatus(NetworkStatusEventArgs args) {
        	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessNetworkStatusMessage(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnSensorStatus(SensorStatusEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessSensorStatusIndicator(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnReadDataAck(ReadConfigEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessReadDatasectorResponse(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnWriteDataAck(WriteConfigEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessWriteDatasectorResponse(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnACTLAck(ActionCtrlEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessAPPCMDResponse(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnBootloaderAck(BootloaderEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessBootloaderResponse(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnParentMsg(ParentMessageEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessParentMsg(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        public void OnProcessData(DataMessageEventArgs args) {
          	for(MineServiceGatewayResponseAccess h : GatewayResponseHandlers) {
        		try {
					h.ProcessDataMessage(args);
				} catch (Exception e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
        	}
        }
        
        ////////////////////////
        //////    Encryption
        ///////////////////////
        
        //internal
        byte[] EncryptPacket(Gateway sensorGateway, byte[] responseBuffer) throws Exception
        {
            try
            {
                byte[] padding = new byte[16 - ((responseBuffer.length - 10) % 16)];
                if (padding.length < 2)
                    padding = new byte[padding.length + 16];
                SecureRandom rnd = new SecureRandom();
                rnd.nextBytes(padding);

                //put the original length in the last 2 bytes of the padding
                System.arraycopy(BitConverter.getTwoBytes(responseBuffer.length), 0, padding, padding.length - 2, 2);

                byte[] encryptBuffer = new byte[responseBuffer.length + padding.length - 10];
                System.arraycopy(responseBuffer, 10, encryptBuffer, 0, responseBuffer.length - 10);
                System.arraycopy(padding, 0, encryptBuffer, responseBuffer.length - 10, padding.length);

                byte[] InitilazationVector = sensorGateway.EncryptionIV;
                //New Random IV helper Last 4 bytes
                byte[] IV = new byte[4];
                rnd.nextBytes(IV);
                //Update ResponseBuffer to new random IV helper
                System.arraycopy(IV, 0, InitilazationVector, 12, 4);
                //Update ResponseBuffer to new random IV helper
                System.arraycopy(IV, 0, responseBuffer, 4, 4);


                EncResult Encrypted = GW_Encryption.Encrypt128(encryptBuffer, sensorGateway.EncryptionKey, InitilazationVector);

                byte[] newBuffer = new byte[Encrypted.result.length + 10];
                System.arraycopy(responseBuffer, 0, newBuffer, 0, 10);
                System.arraycopy(Encrypted.result, 0, newBuffer, 10, Encrypted.result.length);
                return newBuffer;
            }
            catch (Exception ex)
            {
                OnLogException(ex, "TCPServer.EncryptPacket(Gateway sensorGateway, byte[] responseBuffer)");
                throw ex;
               // return responseBuffer;
            }
        }

        //internal
        byte[] DecryptPacket(Gateway sensorGateway, byte[] originalPacket) throws Exception
        {

            	Extensions.debug("Entering DecryptPacket", 30);
            	Extensions.debug(Extensions.hexStr(originalPacket), 30);
                byte[] InitialazationVector = sensorGateway.EncryptionIV;
                System.arraycopy(originalPacket, 4, InitialazationVector, 12, 4);

                byte[] encryptedBuffer = new byte[originalPacket.length - 10];
                System.arraycopy(originalPacket, 10, encryptedBuffer, 0, encryptedBuffer.length);

                byte[] plainText = GW_Encryption.Decrypt128(encryptedBuffer, sensorGateway.EncryptionKey, InitialazationVector);
                int len = BitConverter.twoBytesToInt(plainText, plainText.length - 2);

                byte[] newPacket = new byte[len];
                System.arraycopy(originalPacket, 0, newPacket, 0, 10);
                System.arraycopy(plainText, 0, newPacket, 10, len - 10);

                return newPacket;

/*            }
            catch (Exception ex)
            {
                OnLogException(new Exception(String.format("TCPServer.DecryptPacket(Gateway sensorGateway, byte[] originalPacket) GatewayID: %s - ", sensorGateway.GatewayID)));
                ex.printStackTrace();
                throw (ex);
            }*/
        }

}